/*
 * Test showing that at times a socket close() + exit(0) prevents all the
 * data from reaching the client.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <QCoreApplication>
#include <QByteArray>
#include <QFile>
#include <QTcpSocket>
#include <QTcpServer>
#include <QHostAddress>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/socket.h>
//#include <netinet/in.h>

void
client()
{
	QHostAddress a("192.168.2.1");
	int port = 4004;
	QTcpSocket socket;
	socket.connectToHost(a, port);
	if(socket.waitForConnected())
	{
		// if we get here then we can just copy the output of the child to Apache2
		// the wait will flush all the writes as necessary
		if(!socket.waitForReadyRead())
		{
			fprintf(stderr, "error: waitForReadyRead() timed out.\n");
			exit(1);
		}
		// buffer somewhere to check the output validity
		FILE *f = fopen("/tmp/out.html", "w");
		if(f == NULL)
		{
			fprintf(stderr, "error: could not create /tmp/out.html file.\n");
			exit(1);
		}
		for(;;)
		{
			char buf[64 * 1024];
			qint64 r(socket.read(buf, sizeof(buf)));
			if(r > 0)
			{
				fwrite(buf, r, 1, f);
			}
			else if(r == -1)
			{
				fprintf(stderr, "error: an error occured while read()'ing from the socket.\n");
				break;
			}
			else if(r == 0)
			{
				// no more data (since our socket is blocking this really
				// only happens when no more data is available)
				break;
			}
		}
		fclose(f);
		return;
	}
	fprintf(stderr, "error: client could not connect to server.\n");
	exit(1);
}


void
server()
{
	QTcpServer s;
	QHostAddress a("0.0.0.0");
	if(!s.listen(a, 4004))
	{
		fprintf(stderr, "error: server is not able to listen, port 4004 not available?\n");
		exit(1);
	}
	pid_t pid(0);
	for(;;)
	{
		if(pid > 0)
		{
			int status;
			if(waitpid(pid, &status, WNOHANG) == pid)
			{
				pid = 0;
			}
		}
		s.waitForNewConnection(-1);
		if(pid > 0)
		{
			int status;
			if(waitpid(pid, &status, 0) == pid)
			{
				pid = 0;
			}
		}
		QTcpSocket *socket(s.nextPendingConnection());
		pid = fork();
		if(pid == 0)
		{
			// this makes it work a bit better but we still have the
			// early disconnection
			int optval(1);
			socklen_t optlen(sizeof(optval));
			if(setsockopt(socket->socketDescriptor(), SOL_SOCKET, SO_KEEPALIVE, &optval, optlen) < 0) {
				fprintf(stderr, "warning: could not mark the new connection with keepalive flag.\n");
			}

			// we are the child, write in the socket and close it
			QFile f("example.html");
			f.open(QIODevice::ReadOnly);
			QByteArray buffer(f.readAll());
			f.close();
			socket->write(buffer);
			// the sleep makes it slow, but it doesn't make any difference
			//sleep(2);
			socket->flush();
			sleep(1);
			socket->flush();
			// the read doesn't seem to fix anything
			socket->waitForReadyRead();
			socket->disconnectFromHost();
			delete socket;
			exit(0);
		}
		else {
			if(pid == -1) {
				fprintf(stderr, "error: could not create child process to handle the socket.\n");
			}
			socket->close();
		}
	}
}


int
main(int argc, char *argv[])
{
	if(argc == 1)
	{
		client();
	}
	else if(strcmp(argv[1], "-s") == 0)
	{
		server();
	}
	else
	{
		printf("Usage: %s [-s]\n", argv[0]);
		exit(1);
	}

	exit(0);
}

// vim: ts=4 sw=4
